import numpy as np 
import cv2

STAGE_FIRST_FRAME = 0
STAGE_SECOND_FRAME = 1
STAGE_DEFAULT_FRAME = 2
kMinNumFeature = 1500

lk_params = dict(winSize  = (21, 21), 
				#maxLevel = 3,
             	criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 30, 0.01))

def featureTracking(image_ref, image_cur, px_ref):
	kp2, st, err = cv2.calcOpticalFlowPyrLK(image_ref, image_cur, px_ref, None, **lk_params)  #shape: [k,2] [k,1] [k,1]

	st = st.reshape(st.shape[0])		
	kp1 = px_ref[st == 1]
	kp2 = kp2[st == 1]

	return kp1, kp2


class PinholeCamera:
	def __init__(self, width, height, fx, fy, cx, cy, 
				k1=0.0, k2=0.0, p1=0.0, p2=0.0, k3=0.0):
		self.width = width
		self.height = height
		self.fx = fx
		self.fy = fy
		self.cx = cx
		self.cy = cy
		self.distortion = (abs(k1) > 0.0000001)
		self.d = [k1, k2, p1, p2, k3]


class VisualOdometry:
	def __init__(self, cam,detector):
		self.frame_stage = 0
		self.cam = cam
		self.new_frame = None
		self.last_frame = None
		self.cur_R = None
		self.cur_t = None
		self.px_ref = None
		self.px_cur = None
		self.focal = cam.fx
		self.pp = (cam.cx, cam.cy)
		self.trueX, self.trueY, self.trueZ = 0, 0, 0
		if detector == "FAST":
			self.detector = cv2.FastFeatureDetector_create(threshold=25, nonmaxSuppression=True)
		elif detector == "SIFT":
			self.detector = cv2.xfeatures2d.SIFT_create()


	# px_ref : detected feature points in the "reference" frame
	# px_cur : matched feature points in the "current" frame
	# 

	def exclude_dynamic(self,feature_points,boxes):
		for detected_object in range(0, len(boxes)):
			box = boxes[detected_object]
			left = box[0]
			top = box[1]
			width = box[2]
			height = box[3]
			feature_points = np.array([i for i in feature_points if not((left < i[0] < left+width) and (top < i[1] < top+height))])
		return feature_points 

	def processFirstFrame(self,boxes,yolo = False):
		self.px_ref = self.detector.detect(self.new_frame)
		self.px_ref = np.array([x.pt for x in self.px_ref], dtype=np.float32)
		self.frame_stage = STAGE_SECOND_FRAME
		if yolo:
			self.px_ref = self.exclude_dynamic(self.px_ref,boxes)
		return self.px_ref

	def processSecondFrame(self,boxes,yolo = False):
		if yolo:
			self.px_ref = self.exclude_dynamic(self.px_ref,boxes)
		self.px_ref, self.px_cur = featureTracking(self.last_frame, self.new_frame, self.px_ref)
		E, mask = cv2.findEssentialMat(self.px_cur, self.px_ref, focal=self.focal, pp=self.pp, method=cv2.RANSAC, prob=0.999, threshold=1.0)
		points, self.cur_R, self.cur_t, mask = cv2.recoverPose(E, self.px_cur, self.px_ref, focal=self.focal, pp = self.pp)
		self.frame_stage = STAGE_DEFAULT_FRAME 
		self.px_ref = self.detector.detect(self.new_frame)
		self.px_ref = np.array([x.pt for x in self.px_ref], dtype=np.float32)		# look for new feature points instead of just propagating the previous ones as it was originally
		if yolo:
			self.px_ref = self.exclude_dynamic(self.px_ref,boxes)
		

		# return self.detector.detect(self.new_frame)
		return self.px_ref

	def processFrame(self,boxes,yolo = False):
		if yolo:
			self.px_ref = self.exclude_dynamic(self.px_ref,boxes)
		self.px_ref, self.px_cur = featureTracking(self.last_frame, self.new_frame, self.px_ref)
		E, mask = cv2.findEssentialMat(self.px_cur, self.px_ref, focal=self.focal, pp=self.pp, method=cv2.RANSAC, prob=0.999, threshold=1.0)
		points, R, t, mask = cv2.recoverPose(E, self.px_cur, self.px_ref, focal=self.focal, pp = self.pp)
		# absolute_scale = self.getAbsoluteScale(frame_id)
		# if(absolute_scale > 0.1):
		self.cur_t = self.cur_t + self.cur_R.dot(t) 
		self.cur_R = R.dot(self.cur_R)

		self.px_ref = self.detector.detect(self.new_frame)
		self.px_ref = np.array([x.pt for x in self.px_ref], dtype=np.float32)
		# print(boxes.shape)
		if yolo:
			self.px_ref = self.exclude_dynamic(self.px_ref,boxes)
		
		# return self.detector.detect(self.new_frame)
		return self.px_ref

	def update(self, img,boxes=0,yolo = False):
		assert(img.ndim==2 and img.shape[0]==self.cam.height and img.shape[1]==self.cam.width), "Frame: provided image has not the same size as the camera model or image is not grayscale"
		self.new_frame = img
		if(self.frame_stage == STAGE_DEFAULT_FRAME):
			fp = self.processFrame(boxes, yolo = yolo)
		elif(self.frame_stage == STAGE_SECOND_FRAME):
			fp = self.processSecondFrame(boxes,yolo = yolo)
		elif(self.frame_stage == STAGE_FIRST_FRAME):
			fp = self.processFirstFrame(boxes,yolo = yolo)
		self.last_frame = self.new_frame
		# img=cv2.drawKeypoints(self.new_frame,fp,img)
		# cv2.imshow('',img)
		# cv2.waitKey()
		return fp